#!/usr/bin/env python3

# To write a Python client, we first need to import kgp.  In case
# "kgp.py" is not in the current working directory, adjust your
# PYTHONPATH.
import multiprocessing
multiprocessing.set_start_method("fork")

import kgp

# We will be using a simple evaluation strategy: Prevent the opponent
# from gathering as many stones as possible (we are always playing on
# the south side).  We disregard our own performance, since the total
# number of stones is fixed.  Perhaps that is a bad idea?

def evaluate(state):
    return -state[kgp.NORTH]

# The following procedure implements an actual search procedure.  We
# recursively traverse the search space, using EVALUATE from above to
# find our best/the opponents worst move.
#
# This is the actually interesting part, that you have to improve on.

def search(state, depth, side):
    def child(move):
        if depth <= 0:
            return evaluate(state), move

        after, again = state.sow(side, move)
        if after.is_final():
            return evaluate(after), move
        if again:
            return search(after, depth, side)[0], move
        else:
            return search(after, depth-1, not side)[0], move

    choose = max if side == kgp.SOUTH else min
    return choose((child(move) for move in state.legal_moves(side)),
                  key=lambda ent: ent[0])

# The actual agent is just a generator function, that is to say a
# function that when called will return a generator object.  The
# simplest way to implement this is to use the yield keyword.  In case
# you are not familiar with the keyword, take a look at the Python
# documentation:
#
#       https://docs.python.org/3/reference/expressions.html#yield
#
# The generator function takes a state, i.e. what the board currently
# looks like and yields a series of iterative improvements on what the
# best move should be.
#
# You can generate as many moves as you want, but after the server
# informs the client that the time is over, all following values by
# the generator are ignored.
#
# Note that the agent is invoked in a separate process that will be
# abruptly killed as soon as it's time has come.  By default you
# cannot share resources between invocation.
#
# If you need to synchronise state between searches, use the
# multiprocessing module (specifically multiprocessing.Value or
# multiprocessing.Array might be of interest).

def agent(state):
    yield search(state, 1, kgp.SOUTH)[1] # look ahead one ply
    yield search(state, 2, kgp.SOUTH)[1] # look ahead two plies
    yield search(state, 4, kgp.SOUTH)[1] # look ahead four plies
    yield search(state, 8, kgp.SOUTH)[1] # look ahead eight plies
        
# We can now use the generator function as an agent.  The below
# configuration will check environmental variables that you can set to
# modify the behaviour of this agent.  The relevant options are:
#
# - USE_WEBSOCKET: Connect to the public practice server instead of
#   localhost.  You can set this for the duration of your session, to
#   always connect to the server
#
#   $ USE_WEBSOCKET=t export USE_WEBSOCKET
# 
#   By default the client will connect to your localhost on port 2761.
#   Keep in mind that this is the intended behaviour of your final
#   submission.  
#
#   Note that by default kgp.py requires that the "websocket-client"
#   library (not to be confused with "websockets", that ends with an
#   "s") has to be installed for Python 3, as the public server is only
#   accessible over a websocket connection.
#
# - TOKEN: A random string used to identify your agent.  Think of it
#   as password and username in one.  You can generate a satisfyingly
#   random token in a shell session by running the following command:
# 
#   $ tr -cd 'a-z' </dev/urandom | head -c100
#
# - NAME: Any name for your agent.  This is mainly to make identifying
#   it on the website it easier for you.  The name is public and can
#   be changed whenever the agent connects with the same token.
#
# As with everything here (including the library file kgp.py), you are
# free to change anything.  A more extensive, manual example includes
# some more agent metadata:
#
#     kgp.connect(agent, host    = "wss://kalah.kwarc.info/socket",
#                        token   = "A hopefully random string only I know",
#                        authors = ["Eva Lu Ator", "Ben Bitdiddle"],
#                        name    = "magenta")
#
# This will be sent to the server and used to identify a client over
# multiple connections.  You may leave out the TOKEN keyword, if you
# wish to stay anonymous, in which case your client will not appear on
# the website.

if __name__ == "__main__":
    import os
    host = "wss://kalah.kwarc.info/socket" if os.getenv("USE_WEBSOCKET") else "localhost"
    kgp.connect(agent, host=host,
                token=os.getenv("TOKEN"),
                name=os.getenv("NAME"))
    
